from collections import namedtuple
from typing import Any, Tuple

from rest_framework.exceptions import ValidationError

from django.db.models import Avg, Count, F, Max, Min, Q, Sum
from django.db.models.functions import Trunc

from haupt.db.defs import Models
from haupt.db.query_managers.bookmarks import filter_bookmarks
from haupt.db.query_managers.run import RunsOfflineFilter
from polyaxon.exceptions import PQLException
from polyaxon.schemas import LifeCycle

AGGREGATE_MAP = {
    "avg": Avg,
    "min": Min,
    "max": Max,
    "sum": Sum,
}

GROUPBY_FIELD = {
    "queue": "queue__name",
    "agent": "agent__name",
    "user": "user__username",
    "project": "project__name",
    "component": "component__name",
}


def get_groupby_field(groupby: str):
    if groupby in GROUPBY_FIELD:
        return {f"{groupby}_name": F(GROUPBY_FIELD[groupby])}
    return None


def check_count_groupby(aggregate: str, groupby: str):
    if groupby not in {
        "status",
        "kind",
        "queue",
        "agent",
        "user",
        "project",
        "component",
    }:
        raise ValidationError(
            f"Received an unsupported stats specification: {aggregate} - {groupby}"
        )


def check_aggregate_groupby(aggregate: str, groupby: str):
    if groupby not in {
        "wait_time",
        "duration",
        "cpu",
        "memory",
        "gpu",
        "custom",
        "cost",
    }:
        raise ValidationError(
            f"Received an unsupported stats specification: {aggregate} - {groupby}"
        )


def check_trunc(trunc: str) -> Tuple[str, str]:
    parts = trunc.split(".")
    if len(parts) != 2:
        raise ValidationError(f"Received an unsupported trunc value: {trunc}")

    if parts[1] not in {
        "minute",
        "hour",
        "day",
        "week",
        "month",
    }:
        raise ValidationError(f"Received an unsupported trunc value: {trunc}")

    if parts[0] not in {
        "created_at",
        "started_at",
        "finished_at",
    }:
        raise ValidationError(f"Received an unsupported trunc value: {trunc}")

    return parts[0], parts[1]


def get_aggregation(aggregate: str, groupby: str):
    aggregate = aggregate.split(",")
    results = {}
    for agg in aggregate:
        if agg in AGGREGATE_MAP:
            results["{}__{}".format(groupby, agg)] = AGGREGATE_MAP[agg](groupby)
        elif agg == "total":
            results["{}__total".format(groupby)] = Sum(F(groupby) * F("duration"))
        else:
            raise ValidationError(
                f"Received an unsupported stats specification: "
                f"annotations - {aggregate} - {groupby}. "
                f"Please check the aggregate ({agg})."
            )

    return results


def get_annotation_stats(queryset, aggregate: str, groupby: str):
    if aggregate == "progress":
        return queryset.aggregate(
            done=Count("id", filter=Q(status__in=LifeCycle.DONE_VALUES), distinct=True),
            count=Count("id", distinct=True),
        )
    if aggregate == "count":
        check_count_groupby(aggregate, groupby)
        groupby_field = get_groupby_field(groupby)
        queryset = (
            queryset.values(**groupby_field)
            if groupby_field
            else queryset.values(groupby)
        )
        return queryset.annotate(count=Count("id", distinct=True))

    if aggregate:
        check_aggregate_groupby(aggregate, groupby)
        agg = get_aggregation(aggregate, groupby)
        return queryset.aggregate(**agg)

    raise ValidationError(
        f"Received an unsupported stats specification: annotations - {aggregate} - {groupby}"
    )


def get_series_stats(queryset, aggregate: str, groupby: str, trunc: str):
    if not trunc:
        raise ValidationError(
            f"Received an unsupported stats specification: {aggregate} - {groupby} - {trunc}"
        )
    trunc_field, trunc_frequency = check_trunc(trunc)
    queryset = queryset.annotate(trunc_time=Trunc(trunc_field, trunc_frequency))
    if aggregate == "count":
        check_count_groupby(aggregate, groupby)
        groupby_field = get_groupby_field(groupby)
        queryset = (
            queryset.values("trunc_time", **groupby_field)
            if groupby_field
            else queryset.values(groupby, "trunc_time")
        )
        return queryset.annotate(count=Count("id", distinct=True))
    if aggregate:
        check_aggregate_groupby(aggregate, groupby)
        agg = get_aggregation(aggregate, groupby)
        return queryset.values("trunc_time").annotate(**agg).order_by("trunc_time")

    raise ValidationError(
        f"Received an unsupported stats specification: "
        f"annotations - {aggregate} - {groupby} - {trunc}"
    )


def get_stats(queryset, kind: str, aggregate: str, groupby: str, trunc: str):
    if kind == "annotations":
        return get_annotation_stats(
            queryset=queryset, aggregate=aggregate, groupby=groupby
        )
    elif kind == "series":
        return get_series_stats(
            queryset=queryset, aggregate=aggregate, groupby=groupby, trunc=trunc
        )

    raise ValidationError(f"Received an unsupported kind: {kind}")


class StatsSerializer(
    namedtuple("StatsSerializer", "queryset kind aggregate groupby trunc")
):
    @property
    def data(self):
        if self.queryset is not None:
            return get_stats(
                queryset=self.queryset,
                kind=self.kind,
                aggregate=self.aggregate,
                groupby=self.groupby,
                trunc=self.trunc,
            )
        raise Exception("StatsSerializer was not configured correctly")

    @classmethod
    def filter_queryset(cls, queryset, request, view):
        queryset = filter_bookmarks(queryset, request, view)
        offline_filter = RunsOfflineFilter()
        try:
            queryset = offline_filter.filter_stats(queryset=queryset, request=request)
        except PQLException as e:
            raise ValidationError(
                "Received a wrong join specification. Error: {}".format(e)
            )

        limit = request.query_params.get("limit")
        offset = request.query_params.get("offset")
        if limit:
            try:
                limit = RunsOfflineFilter.positive_int(limit, strict=True)
                offset = RunsOfflineFilter.positive_int(offset or 0, strict=True)
            except (ValueError, TypeError):
                raise ValidationError(
                    "Received a wrong (limit/offset) specification. "
                    "limit: {}, offset: {}".format(limit, offset)
                )
        if limit:
            return Models.Run.all.filter(
                id__in=queryset.values_list("id", flat=True)[offset : offset + limit]
            )
        return queryset


def annotate_statuses(queryset):
    return queryset.annotate(
        count=Count(
            "runs",
            distinct=True,
        ),
        running=Count(
            "runs",
            filter=Q(runs__status__in=LifeCycle.RUNNING_VALUES),
            distinct=True,
        ),
        pending=Count(
            "runs",
            filter=Q(runs__status__in=LifeCycle.ALL_PENDING_VALUES),
            distinct=True,
        ),
        warning=Count(
            "runs",
            filter=Q(runs__status__in=LifeCycle.ALL_WARNING_VALUES),
            distinct=True,
        ),
    )


def annotate_quota(queryset):
    queryset_filter = Q(runs__status__in=LifeCycle.ON_K8S_VALUES)
    return queryset.annotate(
        cost=Sum("runs__cost", filter=queryset_filter),
        custom=Sum("runs__custom", filter=queryset_filter),
        gpu=Sum("runs__gpu", filter=queryset_filter),
        cpu=Sum("runs__cpu", filter=queryset_filter),
        memory=Sum("runs__memory", filter=queryset_filter),
    )


def collect_project_run_count_stats(project: Models.Project):
    data = (
        Models.Run.all.filter(project=project)
        .values("live_state")
        .annotate(run_count=Count("id"))
    )
    return {item["live_state"]: item["run_count"] for item in data}


def collect_project_run_status_stats(project: Models.Project):
    data = (
        Models.Run.objects.filter(project=project)
        .values("status")
        .annotate(run_count=Count("id"))
    )
    return {item["status"]: item["run_count"] for item in data}


def collect_project_run_duration_stats(project: Models.Project):
    data = (
        Models.Run.all.filter(project=project)
        .values("live_state")
        .annotate(sum_duration=Sum("duration"))
    )
    return {item["live_state"]: item["sum_duration"] for item in data}


def collect_project_version_stats(project: Models.Project):
    data = project.versions.values("kind").annotate(kind_count=Count("kind"))
    return {item["kind"]: item["kind_count"] for item in data}


def collect_project_unique_user_stats(project):
    runs_unique_users = list(
        Models.Run.all.filter(project=project)
        .filter(contributors__isnull=False)
        .values_list("contributors__id", flat=True)
        .distinct()
    )
    versions_unique_users = list(
        project.versions.filter(contributors__isnull=False)
        .values_list("contributors__id", flat=True)
        .distinct()
    )
    unique_users = set(runs_unique_users + versions_unique_users)
    if not unique_users:
        return {}
    return {"count": len(unique_users), "ids": list(unique_users)}


def collect_org_project_count_stats(org: Any):
    data = (
        Models.Project.all.filter(owner=org)
        .values("live_state")
        .annotate(project_count=Count("id"))
    )
    return {item["live_state"]: item["project_count"] for item in data}


def collect_org_projects_contributors(org: Any):
    return set(
        list(
            Models.Project.all.filter(owner=org)
            .filter(contributors__isnull=False)
            .values_list("contributors__id", flat=True)
            .distinct()
        )
    )



from haupt import settings
from haupt.proxies.schemas.base import get_config

OPTIONS = r"""
#!/bin/bash
set -e
set -o pipefail

{cmd}
"""


def get_forward_cmd():
    if not settings.PROXIES_CONFIG.has_forward_proxy:
        return

    cmd = None
    if settings.PROXIES_CONFIG.forward_proxy_kind == "transparent":
        cmd = "socat TCP4-LISTEN:8443,reuseaddr,fork TCP:{proxy_host}:{proxy_port}".format(
            proxy_host=settings.PROXIES_CONFIG.forward_proxy_host,
            proxy_port=settings.PROXIES_CONFIG.forward_proxy_port,
        )
    elif (
        settings.PROXIES_CONFIG.forward_proxy_kind is None
        or settings.PROXIES_CONFIG.forward_proxy_kind == "connect"
    ):
        cmd = (
            "socat TCP4-LISTEN:8443,reuseaddr,fork,bind=127.0.0.1 "
            "PROXY:{proxy_host}:{api_host}:{api_port},proxyport={proxy_port}".format(
                api_host=settings.PROXIES_CONFIG.api_host,
                api_port=settings.PROXIES_CONFIG.api_port,
                proxy_host=settings.PROXIES_CONFIG.forward_proxy_host,
                proxy_port=settings.PROXIES_CONFIG.forward_proxy_port,
            )
        )

    if not cmd:
        return

    return get_config(
        options=OPTIONS,
        indent=0,
        cmd=cmd,
    )



import logging
import os

from typing import List, Optional

from clipped.utils.lists import to_list

from haupt import settings
from polyaxon._contexts import paths as ctx_paths
from polyaxon._env_vars.keys import ENV_KEYS_SERVICE, ENV_KEYS_UI_IN_SANDBOX
from polyaxon._services.values import PolyaxonServices

_logger = logging.getLogger("haupt.cli.manage")


def manage(command: str, args: Optional[List[str]] = None):
    from django.core.management import execute_from_command_line

    # Required env var to trigger a management command
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "haupt.polyconf.settings")

    argv = ["manage.py"]
    if command:
        argv.append(command)
    argv += to_list(args, check_none=True)
    execute_from_command_line(argv)


def migrate(
    migrate_tables: bool = False,
    migrate_db: bool = False,
):
    # Required env var to trigger a management command
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "haupt.polyconf.settings")

    if migrate_tables:
        _logger.info("Starting tables migration ...")
        manage("tables")
        _logger.info("Tables were migrated correctly!")

    if migrate_db:
        manage("migrate")
        _logger.info("DB Migration finished!")


def run_manage(command: str, args: Optional[List[str]] = None):
    os.environ[ENV_KEYS_SERVICE] = PolyaxonServices.API
    os.environ[ENV_KEYS_UI_IN_SANDBOX] = "true"
    settings.set_sandbox_config(path=ctx_paths.CONTEXT_ARTIFACTS_ROOT)
    manage(command, args)



from haupt.common import conf
from haupt.common.options.registry import installation

conf.subscribe(installation.PlatformEnvironmentVersion)
conf.subscribe(installation.PlatformVersion)
conf.subscribe(installation.PlatformDist)
conf.subscribe(installation.PlatformHost)
conf.subscribe(installation.ChartVersion)
conf.subscribe(installation.OrganizationKey)



import os

from pathlib import Path

from haupt.schemas.platform_config import PlatformConfig
from polyaxon._config.manager import ConfigManager
from polyaxon._config.spec import ConfigSpec
from polyaxon._env_vars.keys import ENV_KEYS_PLATFORM_CONFIG


class PlatformManager(ConfigManager):
    """Manages platform configuration file."""

    VISIBILITY = ConfigManager.Visibility.GLOBAL
    CONFIG_FILE_NAME = ".platform"
    CONFIG = PlatformConfig

    @classmethod
    def get_config_from_env(cls, **kwargs) -> PlatformConfig:
        def base_directory():
            root = Path(os.path.abspath(kwargs.get("file_path", __file__)))
            root.resolve()
            return root.parent.parent

        root_dir = base_directory()

        config_module = "polyconf"
        config_prefix = os.environ.get("CONFIG_PREFIX")
        if config_prefix:
            config_module = "{}.{}".format(config_prefix, config_module)
        glob_path = cls.get_global_config_path()
        platform_config_path = os.getenv(ENV_KEYS_PLATFORM_CONFIG) or {}
        config_paths = [
            os.environ,
            {
                "POLYAXON_CONFIG_MODULE": config_module,
                "POLYAXON_CONFIG_ROOT_DIR": root_dir,
            },
            ConfigSpec(
                platform_config_path, config_type=".json", check_if_exists=False
            ),
            ConfigSpec(glob_path, config_type=".json", check_if_exists=False),
        ]

        platform_config = cls._CONFIG_READER.read_configs(config_paths)
        return PlatformConfig.from_dict(platform_config.data)



from haupt.common.events.registry import run
from haupt.orchestration import executor

executor.subscribe(run.RunCreatedEvent)
executor.subscribe(run.RunResumedActorEvent)
executor.subscribe(run.RunStoppedActorEvent)
executor.subscribe(run.RunApprovedActorEvent)
executor.subscribe(run.RunNewStatusEvent)
executor.subscribe(run.RunDoneEvent)
executor.subscribe(run.RunDeletedActorEvent)
executor.subscribe(run.RunNewArtifactsEvent)



from haupt import settings
from haupt.proxies.schemas.auth import get_auth_config
from haupt.proxies.schemas.base import clean_config
from haupt.proxies.schemas.dns import get_resolver
from haupt.proxies.schemas.gateway.api import get_api_location_config
from haupt.proxies.schemas.gateway.healthz import get_healthz_location_config
from haupt.proxies.schemas.locations import get_platform_locations_config
from haupt.proxies.schemas.scaffold import get_scaffold_config
from haupt.proxies.schemas.streams.api import get_platform_config
from haupt.proxies.schemas.streams.k8s import get_k8s_root_location_config


def get_base_config(is_platform: bool = True):
    resolver = get_resolver()
    auth = get_auth_config()
    if is_platform:
        api_configs = [
            get_platform_config(),
        ]
        api_location_configs = [
            get_platform_locations_config(),
            get_k8s_root_location_config(),
        ]
        port = settings.PROXIES_CONFIG.api_target_port
    else:
        api_configs = None
        api_location_configs = [
            get_api_location_config(resolver=resolver, auth=auth),
            get_healthz_location_config(),
        ]
        port = settings.PROXIES_CONFIG.gateway_target_port
    config = get_scaffold_config(
        is_proxy=not is_platform,
        port=port,
        use_ssl_config=True,
        use_assets_config=True,
        use_services_configs=True,
        resolver=resolver,
        auth=auth,
        api_configs=api_configs,
        api_location_configs=api_location_configs,
        is_local_streams_service=is_platform,
    )

    return clean_config(config)



import os

import click


@click.command()
@click.argument("component")
@click.option(
    "--path", default="./web", help="Path where the config should be generated."
)
@click.option("--root", help="Absolute root where the configs, default to pwd")
def proxy(component, path, root):
    """Create api proxy."""
    from haupt import settings
    from haupt.proxies.generators import (
        generate_api_conf,
        generate_forward_proxy_cmd,
        generate_gateway_conf,
        generate_streams_conf,
    )
    from haupt.settings import set_proxies_config
    from polyaxon._services.values import PolyaxonServices
    from polyaxon.exceptions import PolyaxonException

    if not root:
        root = os.path.abspath(".")

    set_proxies_config()

    if PolyaxonServices.is_api(component):
        generate_api_conf(path=path, root=root)
    elif PolyaxonServices.is_streams(component):
        generate_streams_conf(path=path, root=root)
        generate_forward_proxy_cmd(path=path)
    elif PolyaxonServices.is_gateway(component):
        generate_gateway_conf(path=path, root=root)
        generate_forward_proxy_cmd(path=path)
    else:
        raise PolyaxonException("Component {} is not recognized".format(component))

    if not os.path.exists(settings.PROXIES_CONFIG.logs_root):
        os.makedirs(settings.PROXIES_CONFIG.logs_root)



from haupt.db.abstracts.project_versions import BaseProjectVersion


class ProjectVersion(BaseProjectVersion):
    class Meta(BaseProjectVersion.Meta):
        app_label = "db"
        db_table = "db_projectversion"



from haupt.proxies.schemas.base import get_config

OPTIONS = """
error_page 500 502 503 504 /static/errors/50x.html;
error_page 401 403 /static/errors/permission.html;
error_page 404 /static/errors/404.html;
"""


def get_error_page_config():
    return get_config(options=OPTIONS, indent=0)



from django.conf import settings

from haupt.db.query_managers.manager import BaseQueryManager
from polyaxon._pql.builder import ArrayCondition, KeysCondition, ValueCondition
from polyaxon._pql.parser import parse_value_operation


class RunEdgeQueryManager(BaseQueryManager):
    NAME = "run_edge"
    FIELDS_ORDERING = (
        "kind",
        "upstream",
        "downstream",
    )
    FIELDS_USE_UUID = {"upstream", "downstream"}
    FIELDS_PROXY = {
        "id": "uuid",
        "uid": "uuid",
    }
    FIELDS_ORDERING_PROXY = {
        "values": {"field": "values", "annotate": True},
    }
    CHECK_ALIVE = False
    PARSERS_BY_FIELD = {
        # upstream
        "upstream": parse_value_operation,
        # downstream
        "downstream": parse_value_operation,
        # Kind
        "kind": parse_value_operation,
        # Status
        "statuses": parse_value_operation,
        # Values
        "values": parse_value_operation,
    }
    CONDITIONS_BY_FIELD = {
        # upstream
        "upstream": ValueCondition,
        # downstream
        "downstream": ValueCondition,
        # Kind
        "kind": ValueCondition,
        # Status
        "statuses": KeysCondition
        if settings.DB_ENGINE_NAME == "sqlite"
        else ArrayCondition,
        # Values
        "values": ValueCondition,
    }



from haupt.db.administration.utils import DiffModelAdmin


class ArtifactAdmin(DiffModelAdmin):
    list_display = (
        "name",
        "kind",
        "state",
    )
    list_display_links = (
        "name",
        "kind",
        "state",
    )
    readonly_fields = DiffModelAdmin.readonly_fields + ("name",)
    fields = ("name", "kind", "created_at", "updated_at")



from typing import Optional

from haupt.db.managers.stages import new_stage
from haupt.db.managers.versions import add_version_contributors
from polyaxon.schemas import V1StageCondition


def create_stage(view, serializer, event_type: Optional[str] = None):
    serializer.is_valid()
    validated_data = serializer.validated_data
    if not validated_data:
        return
    condition = None
    if validated_data.get("condition"):
        condition = V1StageCondition.get_condition(**validated_data.get("condition"))
    if condition:
        new_stage(
            entity=view.version,
            owner_id=view._owner_id,
            user=view.request.user,
            condition=condition,
            event_type=event_type,
        )
        add_version_contributors(view.version, users=[view.request.user])



from haupt import settings
from haupt.proxies.schemas.base import get_config
from haupt.proxies.schemas.urls import (
    get_header_host,
    get_service_proxy,
    get_service_url,
    get_ssl_server_name,
)
from polyaxon.api import AUTH_REQUEST_V1_LOCATION, AUTH_V1_LOCATION

AUTH_OPTIONS = r"""
    auth_request     {auth_api};
    auth_request_set $auth_status $upstream_status;
"""  # noqa


def get_auth_config(auth_api: str = AUTH_REQUEST_V1_LOCATION):
    return get_config(
        options=AUTH_OPTIONS if settings.PROXIES_CONFIG.auth_enabled else "",
        indent=0,
        auth_api=auth_api,
    )


AUTH_LOCATION_CONFIG = r"""
location = {auth_api} {{
    {resolver}
    {ssl_server_name}
    proxy_pass {service};
    proxy_pass_request_body off;
    proxy_set_header Content-Length "";
    proxy_set_header X-Real-IP $remote_addr;
    proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
    proxy_set_header X-Forwarded-Proto $scheme;
    proxy_set_header X-Origin-URI $request_uri;
    proxy_set_header X-Origin-Method $request_method;
    {header_host}
    internal;
}}
"""


def get_auth_location_config(resolver: str, is_local_service: bool = False):
    """Note that this is not used anymore and was deprecated in favor of the new auth-request"""
    if settings.PROXIES_CONFIG.auth_external:
        service = settings.PROXIES_CONFIG.auth_external
    elif is_local_service:
        service = "http://polyaxon"
    else:
        service = get_service_url(
            host=settings.PROXIES_CONFIG.api_host,
            port=settings.PROXIES_CONFIG.api_port,
        )
    if is_local_service or not settings.PROXIES_CONFIG.auth_use_resolver:
        resolver = ""
    header_host = get_header_host(service)
    if settings.PROXIES_CONFIG.has_forward_proxy:
        service = get_service_proxy()
    return get_config(
        options=AUTH_LOCATION_CONFIG if settings.PROXIES_CONFIG.auth_enabled else "",
        indent=0,
        service=service,
        auth_api=AUTH_V1_LOCATION,
        resolver=resolver,
        ssl_server_name=get_ssl_server_name(service),
        header_host=header_host,
    )



from rest_framework import status
from rest_framework.response import Response

from haupt.common import conf
from haupt.common.apis.regex import INSTALLATION_KEY, NAME_KEY, VERSION_KEY
from haupt.common.endpoints.base import BaseEndpoint, RetrieveEndpoint
from haupt.common.options.registry.installation import (
    ORGANIZATION_KEY,
    PLATFORM_DIST,
    PLATFORM_VERSION,
)
from haupt.db.managers.dummy_key import get_dummy_key
from polyaxon._cli.session import get_compatibility


class VersionsInstalledView(BaseEndpoint, RetrieveEndpoint):
    ALLOWED_METHODS = ["GET"]

    def retrieve(self, request, *args, **kwargs):
        data = {
            "key": conf.get(ORGANIZATION_KEY) or get_dummy_key(),
            "version": conf.get(PLATFORM_VERSION),
            "dist": conf.get(PLATFORM_DIST),
        }
        return Response(data)


class VersionsCompatibilityView(BaseEndpoint, RetrieveEndpoint):
    ALLOWED_METHODS = ["GET"]
    CONTEXT_KEYS = (INSTALLATION_KEY, VERSION_KEY, NAME_KEY)

    def get(self, request, *args, **kwargs):
        compatibility = get_compatibility(
            key=self.installation,
            service=self.name,
            version=self.version,
            is_cli=False,
            set_config=False,
        )
        return Response(
            data=compatibility.to_dict() if compatibility else {},
            status=status.HTTP_200_OK,
        )



from typing import List, Tuple

from clipped.utils.imports import import_string

from django.conf import settings


class PermissionMapping:
    _MAPPING = {}

    @classmethod
    def get(cls, permissions: List[str]) -> Tuple:
        return tuple(
            cls._MAPPING[permission]
            for permission in permissions
            if permission in cls._MAPPING
        )


PERMISSIONS_MAPPING = (
    import_string(settings.PERMISSIONS_MAPPING)
    if settings.PERMISSIONS_MAPPING
    else PermissionMapping
)



from haupt.proxies.schemas.base import get_config

OPTIONS = """
client_max_body_size        0;
client_body_buffer_size     50m;
client_body_in_file_only clean;
sendfile on;
"""


def get_buffering_config():
    return get_config(options=OPTIONS, indent=0)



from django.db import models


class StateModel(models.Model):
    state = models.UUIDField(db_index=True)

    class Meta:
        abstract = True


class OptionalStateModel(models.Model):
    state = models.UUIDField(null=True, blank=True, db_index=True)

    class Meta:
        abstract = True



from haupt import settings
from haupt.proxies.schemas.base import get_config

OPTIONS = """
send_timeout {timeout};
keepalive_timeout {timeout};
uwsgi_read_timeout {timeout};
uwsgi_send_timeout {timeout};
client_header_timeout {timeout};
proxy_read_timeout {timeout};
keepalive_requests 10000;
"""


def get_timeout_config():
    return get_config(
        options=OPTIONS, indent=0, timeout=settings.PROXIES_CONFIG.nginx_timeout
    )



from clipped.utils.tz import get_datetime_from_now

from django.conf import settings
from django.db.models import Count, Q

from haupt.background.celeryp.tasks import SchedulerCeleryTasks
from haupt.common import conf, workers
from haupt.common.options.registry.cleaning import (
    CLEANING_INTERVALS_ARCHIVES,
    CLEANING_INTERVALS_DELETION,
)
from haupt.db.defs import Models
from haupt.db.managers.live_state import confirm_delete_runs
from polyaxon.schemas import LifeCycle, LiveState, V1RunKind


class CronsDeletionManager:
    @staticmethod
    def delete_archived_projects():
        last_date = get_datetime_from_now(days=conf.get(CLEANING_INTERVALS_ARCHIVES))
        ids = Models.Project.archived.filter(archived_at__lte=last_date).values_list(
            "id", flat=True
        )
        for _id in ids:
            workers.send(
                SchedulerCeleryTasks.DELETE_ARCHIVED_PROJECT,
                kwargs={"project_id": _id},
            )

    @staticmethod
    def delete_in_progress_projects():
        last_date = get_datetime_from_now(days=conf.get(CLEANING_INTERVALS_DELETION))
        Models.Project.all.filter(
            live_state=LiveState.DELETION_PROGRESSING, deleted_at__lte=last_date
        ).delete()

    @staticmethod
    def delete_in_progress_runs():
        # Delete pipelines without managed runs
        runs = (
            Models.Run.all.filter(
                kind__in={V1RunKind.DAG, V1RunKind.MATRIX, V1RunKind.SCHEDULE},
                live_state=LiveState.DELETION_PROGRESSING,
            )
            .annotate(
                unfinished=Count(
                    "pipeline_runs",
                    filter=~Q(
                        pipeline_runs__status__in=LifeCycle.DONE_VALUES
                        | LifeCycle.PENDING_VALUES
                    ),
                    distinct=True,
                )
            )
            .filter(unfinished=0, deleted_at__isnull=True)
        )
        confirm_delete_runs(runs=runs)

        if settings.HAS_ORG_MANAGEMENT:
            # Delete runs that are not assigned to an agent
            runs = Models.Run.all.filter(
                kind__in={
                    V1RunKind.JOB,
                    V1RunKind.SERVICE,
                    V1RunKind.TUNER,
                    V1RunKind.NOTIFIER,
                },
                live_state=LiveState.DELETION_PROGRESSING,
                agent__isnull=True,
                deleted_at__isnull=True,
            )
            confirm_delete_runs(runs=runs)

        last_date = get_datetime_from_now(days=1)
        Models.Run.all.filter(
            live_state=LiveState.DELETION_PROGRESSING, deleted_at__lte=last_date
        ).delete()

    @staticmethod
    def delete_archived_runs():
        last_date = get_datetime_from_now(days=conf.get(CLEANING_INTERVALS_ARCHIVES))
        ids = Models.Run.archived.filter(
            # We only check values that will not be deleted by the archived projects
            project__live_state=LiveState.LIVE,
            archived_at__lte=last_date,
        ).values_list("id", flat=True)
        for _id in ids:
            workers.send(
                SchedulerCeleryTasks.DELETE_ARCHIVED_RUN,
                kwargs={"run_id": _id},
            )



from clipped.compact.pydantic import ValidationError as PydanticValidationError
from clipped.utils.lists import to_list
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from traceml.artifacts import V1RunArtifact, V1RunArtifacts


def create(view, request, *args, **kwargs):
    if not request.data:
        raise ValidationError("Received no artifacts.")

    data = request.data
    if "artifacts" in data:
        data = data["artifacts"]
        try:
            V1RunArtifacts.from_dict(request.data)
        except PydanticValidationError as e:
            raise ValidationError(e)
    else:  # Backward compatibility with legacy clients
        data = to_list(data)
        try:
            [V1RunArtifact.from_dict(r) for r in data]
        except PydanticValidationError as e:
            raise ValidationError(e)

    view.audit(request, *args, **kwargs, artifacts=data)
    return Response(status=status.HTTP_201_CREATED)



from typing import List, Optional

from django.contrib import admin
from django.urls import include, re_path
from django.views.decorators.csrf import ensure_csrf_cookie

from haupt.common import conf
from haupt.common.apis.index.errors import (
    Handler50xView,
    Handler403View,
    Handler404View,
)
from haupt.common.apis.index.views import IndexView
from haupt.common.options.registry.core import UI_ADMIN_ENABLED
from polyaxon.api import ADMIN_V1, UI_V1


def _handler500(request):
    return Handler50xView.as_view()(request)


handler404 = Handler404View.as_view()
handler403 = Handler403View.as_view()
handler500 = _handler500


def get_ui_urlpatterns(ui_urlpatterns):
    ui_patterns = [
        re_path(pattern, ensure_csrf_cookie(IndexView.as_view()), name="index")
        for pattern in ui_urlpatterns
    ]
    return [
        re_path(r"^$", ensure_csrf_cookie(IndexView.as_view()), name="index"),
        re_path(
            r"^{}$".format(UI_V1), ensure_csrf_cookie(IndexView.as_view()), name="ui"
        ),
        re_path(
            r"^{}/".format(UI_V1), include((ui_patterns, "ui_v1"), namespace="ui_v1")
        ),
    ]


def get_base_health_urlpatterns():
    from haupt.common.apis.index.health import HealthView

    return [
        re_path(r"^healthz/?$", HealthView.as_view(), name="health_check"),
    ]


def get_urlpatterns(
    app_patterns: List, no_healthz: bool = False, ui_urlpatterns: Optional[List] = None
):
    if conf.get(UI_ADMIN_ENABLED):
        app_patterns += [re_path(r"^{}/".format(ADMIN_V1), admin.site.urls)]

    urlpatterns = app_patterns
    if not no_healthz:
        urlpatterns = get_base_health_urlpatterns() + urlpatterns
    if ui_urlpatterns:
        urlpatterns += get_ui_urlpatterns(ui_urlpatterns)

    return urlpatterns



from haupt import settings
from haupt.proxies.schemas.base import get_config
from haupt.proxies.schemas.urls import get_service_url
from polyaxon.api import (
    AUTH_REQUEST_V1_LOCATION,
    EXTERNAL_V1_LOCATION,
    MONITORS_V1_LOCATION,
    REWRITE_EXTERNAL_V1_LOCATION,
    REWRITE_MONITORS_V1_LOCATION,
    REWRITE_SERVICES_V1_LOCATION,
    SERVICES_V1_LOCATION,
)

PLUGIN_OPTIONS = r"""
location ~ /{plugin_name}/proxy/([-_.:\w]+)/(.*) {{
    {auth}
    {resolver}
    rewrite_log on;
    rewrite ^/{plugin_name}/proxy/([-_.:\w]+)/(.*) /{plugin_name}/proxy/$1/$2 break;
    proxy_pass http://$1:{port};
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_hide_header X-Frame-Options;
    proxy_set_header Origin "";
    proxy_buffering off;
}}
"""  # noqa


def get_plugin_location_config(name: str, port: int, resolver: str, auth: str):
    return get_config(
        options=PLUGIN_OPTIONS,
        indent=0,
        plugin_name=name,
        port=port,
        resolver=resolver,
        auth=auth,
    )


def get_plugins_location_config(resolver: str, auth: str, proxy_services=None):
    plugins = []

    if proxy_services:
        for plugin, config in proxy_services.items():
            plugins.append(
                get_plugin_location_config(
                    name=plugin, port=config["port"], resolver=resolver, auth=auth
                )
            )

    return plugins


SERVICES_OPTIONS = r"""
location ~ {app}([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/runs/([-_.:\w]+)/([-_.:\w]+)/(.*) {{
    {auth}
    {resolver}
    proxy_pass http://plx-operation-$4.$1.svc.{dns_custom_cluster}:$5;
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_hide_header X-Frame-Options;
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_buffering off;
}}
"""  # noqa

SERVICES_REWRITE_OPTIONS = r"""
location ~ {app}([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/runs/([-_.:\w]+)/([-_.:\w]+)/(.*) {{
    {auth}
    {resolver}
    rewrite_log on;
    rewrite ^{app}([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/runs/([-_.:\w]+)/([-_.:\w]+)/(.*) /$6 break;
    proxy_pass http://plx-operation-$4.$1.svc.{dns_custom_cluster}:$5;
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_hide_header X-Frame-Options;
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_buffering off;
}}
"""  # noqa

EXTERNAL_OPTIONS = r"""
location ~ {app}([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/runs/([-_.:\w]+)/([-_.:\w]+)/(.*) {{
    {resolver}
    proxy_pass http://plx-operation-$4-ext.$1.svc.{dns_custom_cluster}:$5;
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_hide_header X-Frame-Options;
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_buffering off;
}}
"""  # noqa

EXTERNAL_REWRITE_OPTIONS = r"""
location ~ {app}([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/runs/([-_.:\w]+)/([-_.:\w]+)/(.*) {{
    {resolver}
    rewrite_log on;
    rewrite ^{app}([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/runs/([-_.:\w]+)/([-_.:\w]+)/(.*) /$6 break;
    proxy_pass http://plx-operation-$4-ext.$1.svc.{dns_custom_cluster}:$5;
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_hide_header X-Frame-Options;
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_buffering off;
}}
"""  # noqa


def get_services_location_config(
    resolver: str,
    auth: str,
    rewrite: bool = False,
    external: bool = False,
):
    if external:
        options = EXTERNAL_REWRITE_OPTIONS if rewrite else EXTERNAL_OPTIONS
        app = REWRITE_EXTERNAL_V1_LOCATION if rewrite else EXTERNAL_V1_LOCATION
    else:
        options = SERVICES_REWRITE_OPTIONS if rewrite else SERVICES_OPTIONS
        app = REWRITE_SERVICES_V1_LOCATION if rewrite else SERVICES_V1_LOCATION
    return get_config(
        options=options,
        app=app,
        resolver=resolver,
        auth="" if external else auth,
        dns_custom_cluster=settings.PROXIES_CONFIG.dns_custom_cluster,
    )


SERVICES_MONITORS_OPTIONS = r"""
location ~ {app}([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/runs/([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/(.*) {{
    {auth}
    {resolver}
    proxy_pass http://$5.$1.svc.{dns_custom_cluster}:$6;
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_hide_header X-Frame-Options;
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_buffering off;
}}
"""  # noqa

SERVICES_MONITORS_REWRITE_OPTIONS = r"""
location ~ {app}([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/runs/([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/(.*) {{
    {auth}
    {resolver}
    rewrite_log on;
    rewrite ^{app}([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/runs/([-_.:\w]+)/([-_.:\w]+)/([-_.:\w]+)/(.*) /$7 break;
    proxy_pass http://$5.$1.svc.{dns_custom_cluster}:$6;
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_hide_header X-Frame-Options;
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_buffering off;
}}
"""  # noqa


def get_service_monitors_location_config(
    resolver: str, auth: str, rewrite: bool = False
):
    options = (
        SERVICES_MONITORS_REWRITE_OPTIONS if rewrite else SERVICES_MONITORS_OPTIONS
    )
    app = REWRITE_MONITORS_V1_LOCATION if rewrite else MONITORS_V1_LOCATION
    return get_config(
        options=options,
        app=app,
        resolver=resolver,
        auth=auth,
        dns_custom_cluster=settings.PROXIES_CONFIG.dns_custom_cluster,
    )


def get_services_definitions(resolver: str, auth: str):
    return [
        get_services_location_config(
            resolver=resolver,
            auth=auth,
            rewrite=False,
            external=False,
        ),
        get_services_location_config(
            resolver=resolver,
            auth=auth,
            rewrite=True,
            external=False,
        ),
        get_services_location_config(
            resolver=resolver,
            auth=auth,
            rewrite=False,
            external=True,
        ),
        get_services_location_config(
            resolver=resolver,
            auth=auth,
            rewrite=True,
            external=True,
        ),
        get_service_monitors_location_config(
            resolver=resolver,
            auth=auth,
            rewrite=False,
        ),
        get_service_monitors_location_config(
            resolver=resolver,
            auth=auth,
            rewrite=True,
        ),
    ]


STREAMS_OPTIONS = r"""
location /streams/ {{
    {auth}
    {resolver}
    proxy_pass {service};
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_buffering off;
}}
"""  # noqa


def get_streams_service(is_local_service: bool = False) -> str:
    return (
        "http://polyaxon"
        if is_local_service
        else get_service_url(
            host=settings.PROXIES_CONFIG.streams_host,
            port=settings.PROXIES_CONFIG.streams_port,
        )
    )


def get_streams_location_config(
    resolver: str, auth: str, is_local_service: bool = False
):
    service = get_streams_service(is_local_service)
    # Do not use resolve local streams service
    resolver = "" if is_local_service else resolver
    return get_config(
        options=STREAMS_OPTIONS, resolver=resolver, auth=auth, service=service
    )


K8S_OPTIONS = r"""
location /k8s/ {{
    {auth}
    {resolver}
    proxy_pass {service};
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_buffering off;
}}
"""  # noqa


def get_k8s_location_config(resolver: str, auth: str, is_local_service: bool = False):
    service = get_streams_service(is_local_service)
    # Do not use resolve local streams service
    resolver = "" if is_local_service else resolver
    return get_config(
        options=K8S_OPTIONS, resolver=resolver, auth=auth, service=service
    )


INTERNAL_OPTIONS = r"""
location /internal/ {{
    {resolver}
    proxy_pass {service};
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_buffering off;
}}
"""  # noqa


def get_internal_location_config(resolver: str, is_local_service: bool = False):
    service = get_streams_service(is_local_service)
    # Do not use resolve local streams service
    resolver = "" if is_local_service else resolver
    return get_config(options=INTERNAL_OPTIONS, resolver=resolver, service=service)


AUTH_REQUEST_OPTIONS = """
location = {app} {{
    proxy_pass {service};
    proxy_pass_request_body off;
    proxy_set_header Content-Length "";
    proxy_set_header X-Real-IP $remote_addr;
    proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
    proxy_set_header X-Forwarded-Proto $scheme;
    proxy_set_header X-Origin-URI $request_uri;
    proxy_set_header X-Origin-Method $request_method;
    proxy_set_header Host $http_host;
    proxy_intercept_errors {intercept_errors};
    internal;
}}
"""


def get_auth_request_config():
    return get_config(
        options=AUTH_REQUEST_OPTIONS,
        app=AUTH_REQUEST_V1_LOCATION,
        service=get_streams_service(True),
        indent=0,
        intercept_errors="off",
    )



from rest_framework.urlpatterns import format_suffix_patterns

from django.urls import re_path

from haupt.apis.project_resources.views import runs as runs_views
from haupt.apis.project_resources.views import versions as versions_views
from haupt.common.apis.urls import project_versions, projects

projects_urlpatterns = [
    re_path(
        projects.URLS_PROJECTS_RUNS_TAG,
        runs_views.ProjectRunsTagView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_ARTIFACTS_LINEAGE_V0,
        runs_views.ProjectRunsArtifactsView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_ARTIFACTS_LINEAGE,
        runs_views.ProjectRunsArtifactsView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_STOP,
        runs_views.ProjectRunsStopView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_APPROVE,
        runs_views.ProjectRunsApproveView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_DELETE,
        runs_views.ProjectRunsDeleteView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_SYNC,
        runs_views.ProjectRunsSyncView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_LIST,
        runs_views.ProjectRunsListView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_ARCHIVE,
        runs_views.ProjectRunsArchiveView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_RESTORE,
        runs_views.ProjectRunsRestoreView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_TRANSFER,
        runs_views.ProjectRunsTransferView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_INVALIDATE,
        runs_views.ProjectRunsInvalidateView.as_view(),
    ),
    re_path(
        projects.URLS_PROJECTS_RUNS_BOOKMARK,
        runs_views.ProjectRunsBookmarkView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_COMPONENT_VERSIONS_LIST,
        versions_views.ProjectComponentVersionListView.as_view(),
    ),
    # Component Versions
    re_path(
        project_versions.URLS_PROJECT_COMPONENT_VERSIONS_NAMES,
        versions_views.ProjectComponentVersionNameListView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_COMPONENT_VERSIONS_DETAILS,
        versions_views.ProjectComponentVersionDetailView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_COMPONENT_VERSIONS_TRANSFER,
        versions_views.ProjectComponentVersionTransferView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_COMPONENT_VERSIONS_STAGES,
        versions_views.ProjectComponentVersionStageListView.as_view(),
    ),
    # Model Versions
    re_path(
        project_versions.URLS_PROJECT_MODEL_VERSIONS_LIST,
        versions_views.ProjectModelVersionListView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_MODEL_VERSIONS_NAMES,
        versions_views.ProjectModelVersionNameListView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_MODEL_VERSIONS_DETAILS,
        versions_views.ProjectModelVersionDetailView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_MODEL_VERSIONS_TRANSFER,
        versions_views.ProjectModelVersionTransferView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_MODEL_VERSIONS_STAGES,
        versions_views.ProjectModelVersionStageListView.as_view(),
    ),
    # Artifact Versions
    re_path(
        project_versions.URLS_PROJECT_ARTIFACT_VERSIONS_LIST,
        versions_views.ProjectArtifactVersionListView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_ARTIFACT_VERSIONS_NAMES,
        versions_views.ProjectArtifactVersionNameListView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_ARTIFACT_VERSIONS_DETAILS,
        versions_views.ProjectArtifactVersionDetailView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_ARTIFACT_VERSIONS_TRANSFER,
        versions_views.ProjectArtifactVersionTransferView.as_view(),
    ),
    re_path(
        project_versions.URLS_PROJECT_ARTIFACT_VERSIONS_STAGES,
        versions_views.ProjectArtifactVersionStageListView.as_view(),
    ),
]

# Order is important, because the patterns could swallow other urls
urlpatterns = format_suffix_patterns(projects_urlpatterns)



from haupt.common import auditor
from haupt.common.events.registry import run

auditor.subscribe(run.RunCreatedEvent)
auditor.subscribe(run.RunSyncedActorEvent)
auditor.subscribe(run.RunResumedEvent)
auditor.subscribe(run.RunStoppedEvent)
auditor.subscribe(run.RunSkippedEvent)
auditor.subscribe(run.RunNewStatusEvent)
auditor.subscribe(run.RunNewArtifactsEvent)
auditor.subscribe(run.RunSucceededEvent)
auditor.subscribe(run.RunFailedEvent)
auditor.subscribe(run.RunDoneEvent)
auditor.subscribe(run.RunCreatedActorEvent)
auditor.subscribe(run.RunUpdatedActorEvent)
auditor.subscribe(run.RunDeletedActorEvent)
auditor.subscribe(run.RunViewedActorEvent)
auditor.subscribe(run.RunStoppedActorEvent)
auditor.subscribe(run.RunTransferredActorEvent)
auditor.subscribe(run.RunApprovedActorEvent)
auditor.subscribe(run.RunInvalidatedActorEvent)
auditor.subscribe(run.RunResumedActorEvent)
auditor.subscribe(run.RunRestartedActorEvent)
auditor.subscribe(run.RunSyncedActorEvent)
auditor.subscribe(run.RunCopiedActorEvent)
auditor.subscribe(run.RunSkippedActorEvent)
auditor.subscribe(run.RunStatsActorEvent)



import os

from typing import Dict, Optional, Union

from rest_framework import status

from django.core.handlers.asgi import ASGIRequest
from django.http import FileResponse, HttpResponse

from aiofiles.os import stat as aio_stat
from haupt.common.endpoints.files import FilePathResponse
from polyaxon._services.values import PolyaxonServices


async def _redirect(
    redirect_path: str, is_file: bool = False, additional_headers: Optional[Dict] = None
) -> HttpResponse:
    headers = {"Content-Type": "", "X-Accel-Redirect": redirect_path}
    if additional_headers:
        headers.update(additional_headers)
    if is_file:
        stat_result = await aio_stat(redirect_path)
        headers["X-Content-Length"] = str(stat_result.st_size)
        headers["Content-Disposition"] = 'attachment; filename="{}"'.format(
            os.path.basename(redirect_path)
        )

    return HttpResponse(headers=headers)


async def redirect_file(
    archived_path: str, additional_headers: Optional[Dict] = None
) -> Union[HttpResponse, FileResponse]:
    if not archived_path:
        return HttpResponse(
            content="Artifact not found: filepath={}",
            status=status.HTTP_404_NOT_FOUND,
        )

    if PolyaxonServices.is_sandbox():
        return FilePathResponse(filepath=archived_path)
    return await _redirect(
        redirect_path=archived_path, is_file=True, additional_headers=additional_headers
    )


async def redirect_api(
    redirect_path: str, additional_headers: Optional[Dict] = None
) -> HttpResponse:
    if not redirect_path:
        return HttpResponse(
            content="API not found",
            status=status.HTTP_404_NOT_FOUND,
        )

    return await _redirect(
        redirect_path=redirect_path,
        is_file=False,
        additional_headers=additional_headers,
    )


def inject_auth_header(request: ASGIRequest, headers: Dict) -> Dict:
    auth = request.headers.get("Authorization")
    if auth:
        headers["Authorization"] = auth
    return headers



from haupt.common.options.option import Option, OptionScope, OptionStores

POLYAXON_ENVIRONMENT = "POLYAXON_ENVIRONMENT"
PLATFORM_VERSION = "PLATFORM_VERSION"
PLATFORM_DIST = "PLATFORM_DIST"
PLATFORM_HOST = "PLATFORM_HOST"
CHART_VERSION = "CHART_VERSION"
ORGANIZATION_KEY = "POLYAXON_ORGANIZATION_KEY"

OPTIONS = {
    POLYAXON_ENVIRONMENT,
    PLATFORM_VERSION,
    PLATFORM_DIST,
    PLATFORM_HOST,
    CHART_VERSION,
    ORGANIZATION_KEY,
}


class PlatformEnvironmentVersion(Option):
    key = POLYAXON_ENVIRONMENT
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = True
    is_list = False
    typing = "str"
    store = OptionStores.SETTINGS
    default = None
    options = None


class PlatformVersion(Option):
    key = PLATFORM_VERSION
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = True
    is_list = False
    typing = "str"
    store = OptionStores.SETTINGS
    default = None
    options = None


class PlatformDist(Option):
    key = PLATFORM_DIST
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = True
    is_list = False
    typing = "str"
    store = OptionStores.SETTINGS
    default = None
    options = None


class PlatformHost(Option):
    key = PLATFORM_HOST
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = True
    is_list = False
    store = OptionStores.SETTINGS
    typing = "str"
    default = None
    options = None


class ChartVersion(Option):
    key = CHART_VERSION
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = True
    is_list = False
    typing = "str"
    store = OptionStores.SETTINGS
    default = None
    options = None


class OrganizationKey(Option):
    key = ORGANIZATION_KEY
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = True
    is_list = False
    typing = "str"
    store = OptionStores.ENV
    default = None
    options = None



from haupt.proxies.schemas.base import get_config

OPTIONS = """
gzip                        on;
gzip_disable                "msie6";
gzip_types                  *;
gzip_proxied                any;
"""


def get_gzip_config():
    return get_config(options=OPTIONS, indent=0)



from django.urls import include, re_path

from haupt import settings
from haupt.common.apis.index import get_urlpatterns, handler403, handler404, handler500
from haupt.common.apis.regex import OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
from haupt.streams.endpoints.artifacts import artifacts_routes
from haupt.streams.endpoints.auth_request import auth_request_routes
from haupt.streams.endpoints.base import base_health_route
from haupt.streams.endpoints.events import events_routes
from haupt.streams.endpoints.k8s import k8s_routes
from haupt.streams.endpoints.local_sandbox import local_sandbox_routes
from haupt.streams.endpoints.logs import internal_logs_routes, logs_routes
from haupt.streams.endpoints.notifications import notifications_routes
from polyaxon.api import API_V1, AUTH_REQUEST_V1, INTERNAL_V1, STREAMS_V1

sandbox_is_local = settings.SANDBOX_CONFIG and settings.SANDBOX_CONFIG.is_local

streams_routes = (
    logs_routes + k8s_routes + notifications_routes + artifacts_routes + events_routes
)

app_urlpatterns = []

if sandbox_is_local:
    app_urlpatterns += [
        re_path(
            r"^{}/".format(API_V1),
            include((local_sandbox_routes, "local-v1"), namespace="local-v1"),
        ),
    ]

app_urlpatterns += [
    re_path(
        r"^{}/".format(INTERNAL_V1),
        include((internal_logs_routes, "internal-v1"), namespace="internal-v1"),
    ),
    re_path(
        r"^{}/".format(AUTH_REQUEST_V1),
        include((auth_request_routes, "auth-request-v1"), namespace="auth-request-v1"),
    ),
    re_path(
        r"^{}/".format(STREAMS_V1),
        include((streams_routes, "streams-v1"), namespace="streams-v1"),
    ),
    base_health_route,
]

# UI
if sandbox_is_local:
    projects_urls = "{}/{}".format(OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN)
    orgs_urls = "orgs/{}".format(OWNER_NAME_PATTERN)
    ui_urlpatterns = [
        r"^$",
        r"^{}/?".format(orgs_urls),
        r"^{}/projects/?".format(orgs_urls),
        r"^{}/?$".format(projects_urls),
        r"^{}/runs.*/?".format(projects_urls),
        r"^{}/jobs.*/?".format(projects_urls),
        r"^{}/cli.*/?".format(projects_urls),
    ]
else:
    ui_urlpatterns = None

handler404 = handler404
handler403 = handler403
handler500 = handler500
urlpatterns = get_urlpatterns(
    app_patterns=app_urlpatterns, no_healthz=True, ui_urlpatterns=ui_urlpatterns
)



from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional
from typing_extensions import Literal

from clipped.compact.pydantic import Extra, Field, validator
from clipped.utils.logging import DEFAULT_LOGS_ROOT

from haupt import pkg
from polyaxon._config.parser import ConfigParser
from polyaxon._env_vars.keys import (
    ENV_KEYS_ARCHIVES_ROOT,
    ENV_KEYS_ARTIFACTS_ROOT,
    ENV_KEYS_DEBUG,
    ENV_KEYS_ENVIRONMENT,
    ENV_KEYS_K8S_NAMESPACE,
    ENV_KEYS_LOG_LEVEL,
    ENV_KEYS_LOGS_ROOT,
    ENV_KEYS_MAX_CONCURRENCY,
    ENV_KEYS_PLATFORM_HOST,
    ENV_KEYS_SECRET_KEY,
    ENV_KEYS_SERVICE,
    ENV_KEYS_STATIC_ROOT,
    ENV_KEYS_STATIC_URL,
    ENV_KEYS_TIME_ZONE,
    ENV_KEYS_UI_ADMIN_ENABLED,
    ENV_KEYS_UI_ASSETS_VERSION,
    ENV_KEYS_UI_BASE_URL,
    ENV_KEYS_UI_ENABLED,
    ENV_KEYS_UI_IN_SANDBOX,
    ENV_KEYS_UI_OFFLINE,
)
from polyaxon._k8s.namespace import DEFAULT_NAMESPACE
from polyaxon._schemas.base import BaseSchemaModel
from polyaxon.exceptions import PolyaxonSchemaError

if TYPE_CHECKING:
    from clipped.compact.pydantic import ModelField


class PlatformConfig(BaseSchemaModel):
    _IDENTIFIER = "app"

    env: Optional[str] = Field(alias=ENV_KEYS_ENVIRONMENT, default="local")
    config_module: Optional[str] = Field(
        alias="POLYAXON_CONFIG_MODULE", default="polyconf"
    )
    root_dir: Optional[Path] = Field(alias="POLYAXON_CONFIG_ROOT_DIR")
    service: Optional[str] = Field(alias=ENV_KEYS_SERVICE)
    is_debug_mode: Optional[bool] = Field(alias=ENV_KEYS_DEBUG, default=False)
    namespace: Optional[str] = Field(
        alias=ENV_KEYS_K8S_NAMESPACE, default=DEFAULT_NAMESPACE
    )
    logs_root: Optional[str] = Field(
        alias=ENV_KEYS_LOGS_ROOT, default=DEFAULT_LOGS_ROOT
    )
    log_level: Optional[str] = Field(alias=ENV_KEYS_LOG_LEVEL, default="WARNING")
    timezone: Optional[str] = Field(alias=ENV_KEYS_TIME_ZONE, default="UTC")
    scheduler_enabled: Optional[bool] = Field(
        alias="POLYAXON_SCHEDULER_ENABLED", default=False
    )
    chart_version: Optional[str] = Field(
        alias="POLYAXON_CHART_VERSION", default=pkg.VERSION
    )
    redis_protocol: Optional[str] = Field(
        alias="POLYAXON_REDIS_PROTOCOL", default="redis"
    )
    redis_password: Optional[str] = Field(alias="POLYAXON_REDIS_PASSWORD")  # secret
    redis_sessions_url: Optional[str] = Field(alias="POLYAXON_REDIS_SESSIONS_URL")
    redis_heartbeat_url: Optional[str] = Field(alias="POLYAXON_REDIS_HEARTBEAT_URL")
    admin_name: Optional[str] = Field(alias="POLYAXON_ADMIN_NAME")
    admin_mail: Optional[str] = Field(alias="POLYAXON_ADMIN_MAIL")
    extra_apps: Optional[List[str]] = Field(alias="POLYAXON_EXTRA_APPS")
    media_root: Optional[str] = Field(alias="POLYAXON_MEDIA_ROOT", default="")
    media_url: Optional[str] = Field(alias="POLYAXON_MEDIA_URL", default="")
    static_root: Optional[str] = Field(alias=ENV_KEYS_STATIC_ROOT)
    static_url: Optional[str] = Field(alias=ENV_KEYS_STATIC_URL)
    artifacts_root: Optional[str] = Field(alias=ENV_KEYS_ARTIFACTS_ROOT)
    archives_root: Optional[str] = Field(alias=ENV_KEYS_ARCHIVES_ROOT)
    max_concurrency: Optional[int] = Field(alias=ENV_KEYS_MAX_CONCURRENCY, default=50)
    broker_backend: Optional[Literal["redis", "rabbitmq"]] = Field(
        alias="POLYAXON_BROKER_BACKEND"
    )
    celery_redis_broker_url: Optional[str] = Field(
        alias="POLYAXON_REDIS_CELERY_BROKER_URL"
    )
    celery_amqp_broker_url: Optional[str] = Field(alias="POLYAXON_AMQP_URL")
    celery_amqp_user: Optional[str] = Field(alias="POLYAXON_RABBITMQ_USER")
    celery_amqp_password: Optional[str] = Field(
        alias="POLYAXON_RABBITMQ_PASSWORD"
    )  # secret
    celery_task_track_started: Optional[bool] = Field(
        alias="POLYAXON_CELERY_TASK_TRACK_STARTED", default=True
    )
    celery_broker_pool_limit: Optional[int] = Field(
        alias="POLYAXON_CELERY_BROKER_POOL_LIMIT", default=100
    )
    celery_confirm_publish: Optional[bool] = Field(
        alias="POLYAXON_CELERY_CONFIRM_PUBLISH", default=True
    )
    celery_result_backend: Optional[str] = Field(
        alias="POLYAXON_REDIS_CELERY_RESULT_BACKEND_URL"
    )
    celery_worker_prefetch_multiplier: Optional[int] = Field(
        alias="POLYAXON_CELERY_WORKER_PREFETCH_MULTIPLIER", default=4
    )
    celery_task_always_eager: Optional[bool] = Field(
        alias="POLYAXON_CELERY_TASK_ALWAYS_EAGER", default=False
    )
    celery_hard_time_limit_delay: Optional[int] = Field(
        alias="POLYAXON_CELERY_HARD_TIME_LIMIT_DELAY", default=180
    )
    celery_worker_max_tasks_per_child: Optional[int] = Field(
        alias="POLYAXON_CELERY_WORKER_MAX_TASKS_PER_CHILD", default=100
    )
    celery_worker_max_memory_per_child: Optional[int] = Field(
        alias="POLYAXON_CELERY_WORKER_MAX_MEMORY_PER_CHILD", default=400000
    )
    intervals_operations_default_retry_delay: Optional[int] = Field(
        alias="POLYAXON_INTERVALS_OPERATIONS_DEFAULT_RETRY_DELAY", default=60
    )
    intervals_operations_max_retry_delay: Optional[int] = Field(
        alias="POLYAXON_INTERVALS_OPERATIONS_MAX_RETRY_DELAY", default=3600
    )
    intervals_runs_scheduler: Optional[int] = Field(
        alias="POLYAXON_INTERVALS_RUNS_SCHEDULER", default=30
    )
    intervals_heartbeat_check: Optional[int] = Field(
        alias="POLYAXON_INTERVALS_HEARTBEAT_CHECK", default=60
    )
    intervals_stats_calculation: Optional[int] = Field(
        alias="POLYAXON_INTERVALS_STATS_CALCULATION", default=60 * 5
    )
    intervals_clean_activity_logs: Optional[int] = Field(
        alias="POLYAXON_INTERVALS_CLEAN_ACTIVITY_LOGS", default=60
    )
    intervals_clean_support_access: Optional[int] = Field(
        alias="POLYAXON_INTERVALS_CLEAN_SUPPORT_ACCESS", default=60
    )
    intervals_clean_notifications: Optional[int] = Field(
        alias="POLYAXON_INTERVALS_CLEAN_NOTIFICATIONS", default=60
    )
    intervals_delete_archived: Optional[int] = Field(
        alias="POLYAXON_INTERVALS_DELETE_ARCHIVED", default=60
    )

    internal_exchange: Optional[str] = Field(
        alias="POLYAXON_INTERNAL_EXCHANGE", default="internal"
    )
    ttl_heartbeat: Optional[int] = Field(
        alias="POLYAXON_TTL_HEARTBEAT", default=60 * 30
    )
    min_artifacts_deletion_timedelta: Optional[int] = Field(
        alias="POLYAXON_MIN_ARTIFACTS_DELETION_TIMEDELTA", default=80
    )
    db_engine_name: Optional[Literal["sqlite", "pgsql"]] = Field(
        alias="POLYAXON_DB_ENGINE", default="sqlite"
    )
    db_name: Optional[str] = Field(alias="POLYAXON_DB_NAME")
    db_user: Optional[str] = Field(alias="POLYAXON_DB_USER")
    db_password: Optional[str] = Field(alias="POLYAXON_DB_PASSWORD")  # secret
    db_host: Optional[str] = Field(alias="POLYAXON_DB_HOST")
    db_port: Optional[str] = Field(alias="POLYAXON_DB_PORT")
    db_conn_max_age: Optional[int] = Field(alias="POLYAXON_DB_CONN_MAX_AGE", default=0)
    db_options: Optional[Dict] = Field(alias="POLYAXON_DB_OPTIONS")
    cors_allowed_origins: Optional[List[str]] = Field(
        alias="POLYAXON_CORS_ALLOWED_ORIGINS", default=[]
    )
    ssl_enabled: Optional[bool] = Field(alias="POLYAXON_SSL_ENABLED", default=False)
    ssl_redirect_enabled: Optional[bool] = Field(
        alias="POLYAXON_SSL_REDIRECT_ENABLED", default=False
    )
    encryption_key: Optional[str] = Field(alias="POLYAXON_ENCRYPTION_KEY")
    encryption_secret: Optional[str] = Field(
        alias="POLYAXON_ENCRYPTION_SECRET"
    )  # secret
    encryption_backend: Optional[str] = Field(alias="POLYAXON_ENCRYPTION_BACKEND")
    secret_key: Optional[str] = Field(
        alias=ENV_KEYS_SECRET_KEY, default="default-secret"
    )  # secret
    secret_internal_token: Optional[str] = Field(
        alias="POLYAXON_SECRET_INTERNAL_TOKEN", default="default-secret"
    )  # secret
    platform_host: Optional[str] = Field(alias=ENV_KEYS_PLATFORM_HOST)
    allowed_hosts: Optional[List[str]] = Field(
        alias="POLYAXON_ALLOWED_HOSTS", default=["*"]
    )
    allowed_versions: Optional[List[str]] = Field(
        alias="POLYAXON_ALLOWED_VERSIONS", default=[]
    )
    acc_password_length: Optional[int] = Field(
        alias="POLYAXON_PASSWORD_LENGTH", default=6
    )
    acc_activation_days: Optional[int] = Field(
        alias="POLYAXON_ACCOUNT_ACTIVATION_DAYS", default=4
    )
    pod_ip: Optional[str] = Field(alias="POLYAXON_POD_IP")
    host_ip: Optional[str] = Field(alias="POLYAXON_HOST_IP")
    all_user_organization: Optional[bool] = Field(
        alias="POLYAXON_ALLOW_USER_ORGANIZATION", default=False
    )
    frontend_debug: Optional[bool] = Field(
        alias="POLYAXON_FRONTEND_DEBUG", default=False
    )
    template_debug: Optional[bool] = Field(alias="DJANGO_TEMPLATE_DEBUG")
    email_from_email: Optional[str] = Field(
        alias="POLYAXON_EMAIL_FROM", default="<Polyaxon>"
    )
    email_host: Optional[str] = Field(alias="POLYAXON_EMAIL_HOST", default="localhost")
    email_port: Optional[int] = Field(alias="POLYAXON_EMAIL_PORT", default=25)
    email_user: Optional[str] = Field(alias="POLYAXON_EMAIL_HOST_USER", default="")
    email_password: Optional[str] = Field(
        alias="POLYAXON_EMAIL_HOST_PASSWORD", default=""
    )
    email_subject_prefix: Optional[str] = Field(
        alias="POLYAXON_EMAIL_SUBJECT_PREFIX", default="[Polyaxon]"
    )
    email_use_tls: Optional[bool] = Field(alias="POLYAXON_EMAIL_USE_TLS", default=False)
    email_backend: Optional[str] = Field(
        alias="POLYAXON_EMAIL_BACKEND",
        default="django.core.mail.backends.console.EmailBackend",
    )
    rest_throttle_rates_impersonate: Optional[int] = Field(
        alias="POLYAXON_THROTTLE_RATES_IMPERSONATE", default=500
    )
    rest_throttle_rates_auth: Optional[int] = Field(
        alias="POLYAXON_THROTTLE_RATES_AUTH", default=60
    )
    rest_throttle_rates_user: Optional[int] = Field(
        alias="POLYAXON_THROTTLE_RATES_USER", default=240
    )
    rest_throttle_rates_agent: Optional[int] = Field(
        alias="POLYAXON_THROTTLE_RATES_AGENT", default=500
    )
    rest_throttle_rates_run_status: Optional[int] = Field(
        alias="POLYAXON_THROTTLE_RATES_RUN_STATUS", default=1000
    )
    rest_throttle_rates_run_lineage: Optional[int] = Field(
        alias="POLYAXON_THROTTLE_RATES_RUN_LINEAGE", default=1000
    )
    rest_throttle_rates_run: Optional[int] = Field(
        alias="POLYAXON_THROTTLE_RATES_RUN", default=15
    )
    rest_throttle_rates_anon: Optional[int] = Field(
        alias="POLYAXON_THROTTLE_RATES_ANON", default=20
    )
    rest_throttle_rates_checks: Optional[int] = Field(
        alias="POLYAXON_THROTTLE_RATES_CHECKS", default=10
    )
    services_analytics_backend: Optional[str] = Field(
        alias="POLYAXON_ANALYTICS_BACKEND"
    )
    services_analytics_options: Optional[Dict] = Field(
        alias="POLYAXON_ANALYTICS_OPTIONS", default={}
    )
    services_transactions_backend: Optional[str] = Field(
        alias="POLYAXON_TRANSACTIONS_BACKEND"
    )
    services_transactions_options: Optional[Dict] = Field(
        alias="POLYAXON_TRANSACTIONS_OPTIONS", default={}
    )
    services_metrics_backend: Optional[str] = Field(alias="POLYAXON_METRICS_BACKEND")
    services_metrics_options: Optional[Dict] = Field(
        alias="POLYAXON_METRICS_OPTIONS", default={}
    )
    services_errors_backend: Optional[str] = Field(alias="POLYAXON_ERRORS_BACKEND")
    services_errors_options: Optional[Dict] = Field(
        alias="POLYAXON_ERRORS_OPTIONS", default={}
    )
    auth_github_options: Optional[Dict] = Field(
        alias="POLYAXON_AUTH_GITHUB_OPTIONS", default={}
    )
    auth_gitlab_options: Optional[Dict] = Field(
        alias="POLYAXON_AUTH_GITLAB_OPTIONS", default={}
    )
    auth_bitbucket_options: Optional[Dict] = Field(
        alias="POLYAXON_AUTH_BITBUCKET_OPTIONS", default={}
    )
    auth_google_options: Optional[Dict] = Field(
        alias="POLYAXON_AUTH_GOOGLE_OPTIONS", default={}
    )
    auth_azure_options: Optional[Dict] = Field(
        alias="POLYAXON_AUTH_AZURE_OPTIONS", default={}
    )
    auth_saml_options: Optional[Dict] = Field(
        alias="POLYAXON_AUTH_SAML_OPTIONS", default={}
    )
    ui_in_sandbox: Optional[bool] = Field(alias=ENV_KEYS_UI_IN_SANDBOX, default=False)
    ui_admin_enabled: Optional[bool] = Field(
        alias=ENV_KEYS_UI_ADMIN_ENABLED, default=False
    )
    ui_base_url: Optional[str] = Field(alias=ENV_KEYS_UI_BASE_URL)
    ui_assets_version: Optional[str] = Field(
        alias=ENV_KEYS_UI_ASSETS_VERSION, default=""
    )
    ui_offline: Optional[bool] = Field(alias=ENV_KEYS_UI_OFFLINE, default=False)
    ui_enabled: Optional[bool] = Field(alias=ENV_KEYS_UI_ENABLED, default=True)

    class Config:
        extra = Extra.ignore

    @validator(
        "extra_apps",
        "cors_allowed_origins",
        "allowed_hosts",
        "allowed_versions",
        pre=True,
    )
    def validate_str_list(cls, v, field: "ModelField"):
        if not isinstance(v, str):
            return v
        try:
            return ConfigParser.parse(List)(
                key=field.name,
                value=v,
                is_optional=True,
            )
        except PolyaxonSchemaError as e:
            raise ValueError("Received an invalid {} `{}`".format(field.name, v)) from e

    @validator(
        "db_options",
        "services_analytics_options",
        "services_transactions_options",
        "services_metrics_options",
        "services_errors_options",
        "auth_github_options",
        "auth_gitlab_options",
        "auth_bitbucket_options",
        "auth_google_options",
        "auth_azure_options",
        "auth_saml_options",
        pre=True,
    )
    def validate_json_fields(cls, v, field: "ModelField"):
        if not isinstance(v, str):
            return v
        try:
            return ConfigParser.parse(Dict)(
                key=field.name,
                value=v,
                is_optional=True,
            )
        except PolyaxonSchemaError as e:
            raise ValueError("Received an invalid {} `{}`".format(field.name, v)) from e

    @validator("log_level", always=True, pre=True)
    def validate_log_level(cls, v):
        if v:
            return v.upper()
        return v

    @property
    def is_sqlite_db_engine(self) -> bool:
        return self.db_engine_name == "sqlite"

    @property
    def is_pgsql_db_engine(self) -> bool:
        return self.db_engine_name == "pgsql"

    @property
    def is_streams_service(self) -> bool:
        return self.service == "streams"

    @property
    def is_monolith_service(self) -> bool:
        return self.service == "monolith"

    @property
    def is_api_service(self) -> bool:
        return self.service == "api" or self.is_monolith_service

    @property
    def is_scheduler_service(self) -> bool:
        return self.service == "scheduler" or self.is_monolith_service

    @property
    def is_test_env(self) -> bool:
        return self.env == "test"

    @property
    def is_local_env(self) -> bool:
        if self.env == "local":
            return True
        return False

    @property
    def is_staging_env(self) -> bool:
        if self.env == "staging":
            return True
        return False

    @property
    def is_production_env(self) -> bool:
        if self.env == "production":
            return True
        return False

    @property
    def is_redis_broker(self):
        return self.broker_backend == "redis"

    @property
    def is_rabbitmq_broker(self):
        return self.broker_backend == "rabbitmq"

    def get_redis_url(self, redis_url) -> str:
        if self.redis_password:
            redis_url = ":{}@{}".format(self.redis_password, redis_url)
        return "{}://{}".format(self.redis_protocol, redis_url)

    def get_broker_url(self) -> str:
        if self.is_redis_broker:
            return self.get_redis_url(self.celery_redis_broker_url)
        if self.is_rabbitmq_broker:
            if self.celery_amqp_user and self.celery_amqp_password:
                return "amqp://{user}:{password}@{url}".format(
                    user=self.celery_amqp_user,
                    password=self.celery_amqp_password,
                    url=self.celery_amqp_broker_url,
                )

            return "amqp://{url}".format(url=self.celery_amqp_broker_url)



# isort: skip_file

from django.apps import AppConfig


class StreamsConfig(AppConfig):
    name = "haupt.streams"
    verbose_name = "Streams"

    def ready(self):
        from haupt.common import conf

        conf.validate_and_setup()

        import haupt.common.options.conf_subscriptions  # noqa



from clipped.utils.enums import get_enum_value

from django.http import HttpRequest

from polyaxon._services.headers import PolyaxonServiceHeaders

try:
    from rest_framework import HTTP_HEADER_ENCODING
except ImportError:
    raise ImportError("This module depends on django rest.")


POLYAXON_HEADERS_USER_ID = "X_POLYAXON_USER_ID"
POLYAXON_HEADERS_PUBLIC_ONLY = "X_POLYAXON_PUBLIC_ONLY"


def get_header(request: HttpRequest, header_key: str, as_bytestring: bool = False):
    """Return request's 'X_POLYAXON_...:' header, as a str or bytestring.

    Hide some test client ickyness where the header can be unicode.
    """
    header = request.META.get("HTTP_{}".format(get_enum_value(header_key)), "")
    if isinstance(header, str) and as_bytestring:
        # Work around django test client oddness
        header = header.encode(HTTP_HEADER_ENCODING)
    return header


def get_service_header(request: HttpRequest):
    """Return request's 'X_POLYAXON_SERVICE:' header, as a bytestring."""
    return get_header(
        request=request, header_key=PolyaxonServiceHeaders.SERVICE, as_bytestring=True
    )


def get_internal_header(request: HttpRequest) -> str:
    """
    Return request's 'X_POLYAXON_INTERNAL:' header, as a bytestring.
    """
    return get_header(
        request=request, header_key=PolyaxonServiceHeaders.INTERNAL, as_bytestring=True
    )


def get_original_uri_header(request: HttpRequest) -> str:
    """
    Return request's 'X_ORIGIN_URI:' header, as a str.
    """
    return get_header(request=request, header_key="X_ORIGIN_URI", as_bytestring=False)


def get_original_method_header(request: HttpRequest) -> str:
    """
    Return request's 'X_ORIGIN_METHOD:' header, as a str.
    """
    return get_header(
        request=request, header_key="X_ORIGIN_METHOD", as_bytestring=False
    )


def get_user_agent_header(request: HttpRequest) -> str:
    """
    Return request's 'USER_AGENT:' header, as a str.
    """
    return get_header(request=request, header_key="USER_AGENT", as_bytestring=False)


def get_referer_header(request: HttpRequest) -> str:
    """
    Return request's 'REFERER:' header, as a str.
    """
    return get_header(request=request, header_key="REFERER", as_bytestring=False)


def get_authorization_header(request: HttpRequest) -> str:
    """
    Return request's 'AUTHORIZATION:' header, as a str.
    """
    return get_header(request=request, header_key="AUTHORIZATION", as_bytestring=False)


def get_remote_client_ip(request: HttpRequest) -> str:
    x_forwarded_for = get_header(
        request=request, header_key="X_FORWARDED_FOR", as_bytestring=False
    )
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip



import os

from haupt.schemas.proxies_config import ProxiesConfig
from polyaxon._config.manager import ConfigManager


class ProxiesManager(ConfigManager):
    """Manages proxies configuration file."""

    VISIBILITY = ConfigManager.Visibility.GLOBAL
    CONFIG_FILE_NAME = ".proxies"
    CONFIG = ProxiesConfig

    @classmethod
    def get_config_from_env(cls, **kwargs) -> ProxiesConfig:
        config_paths = [os.environ, {"dummy": "dummy"}]

        proxy_config = cls._CONFIG_READER.read_configs(config_paths)
        return ProxiesConfig.from_dict(proxy_config.data)



from typing import Dict, Optional

from django.core.exceptions import BadRequest
from django.http import HttpRequest

from haupt.common import conf
from haupt.common.authentication.internal import get_internal_auth
from haupt.common.options.registry import core


def validate_methods(request: HttpRequest, methods: Optional[Dict] = None):
    if methods and request.method not in methods:
        raise BadRequest("Request method not allowed by this endpoint.")


def validate_internal_auth(request: HttpRequest):
    data = get_internal_auth(request=request)
    if data is None:
        raise BadRequest("Request requires an authentication data.")

    if data[1] != conf.get(core.SECRET_INTERNAL_TOKEN):
        raise BadRequest("Request requires a valid authentication token.")



from typing import Any

from rest_framework.authentication import BasicAuthentication

from django.contrib.auth import get_user_model
from django.http import HttpRequest

from polyaxon._services.auth import AuthenticationError, AuthenticationTypes
from polyaxon._services.values import PolyaxonServices


class PolyaxonAuthentication(BasicAuthentication):
    keyword = None

    def __init__(self) -> None:
        if self.keyword not in AuthenticationTypes.VALUES:
            raise AuthenticationError(
                "Authentication bad configuration, "
                "the keyword `{}` is not supported.".format(self.keyword)
            )

    def authenticate_header(self, request: HttpRequest) -> str:
        return self.keyword


class BaseUser:
    NAME = ""

    def __init__(self, service: str):
        if service not in PolyaxonServices.to_set():
            raise ValueError("Received an unrecognized {} service.".format(self.NAME))
        self.pk = -1
        self.id = -1
        self.is_authenticated = True
        self.is_active = True
        self.is_anonymous = False
        self.is_staff = False
        self.is_superuser = False
        self.service = service
        if not self.username:
            raise ValueError("username required")


def is_user(user: Any) -> bool:
    return isinstance(user, get_user_model())


def is_normal_user(user: Any) -> bool:
    return (
        is_user(user)
        and user.is_authenticated
        and not (user.is_staff or user.is_superuser)
    )



from haupt import settings
from haupt.proxies.schemas.auth import get_auth_config
from haupt.proxies.schemas.base import clean_config
from haupt.proxies.schemas.dns import get_resolver
from haupt.proxies.schemas.gateway.api import get_api_location_config
from haupt.proxies.schemas.locations import get_streams_locations_config
from haupt.proxies.schemas.scaffold import get_scaffold_config
from haupt.proxies.schemas.streams.api import get_api_config
from haupt.proxies.schemas.streams.k8s import get_k8s_root_location_config
from polyaxon.api import HEALTHZ_LOCATION


def get_base_config(is_gateway: bool = True):
    resolver = get_resolver()
    auth = get_auth_config()
    api_location_configs = [
        get_api_config(path="= {}".format(HEALTHZ_LOCATION), intercept_errors="off"),
        get_streams_locations_config(),
        get_k8s_root_location_config(),
    ]
    if is_gateway:
        api_location_configs.append(
            get_api_location_config(resolver=resolver, auth=auth)
        )
    config = get_scaffold_config(
        is_proxy=False,
        port=settings.PROXIES_CONFIG.streams_target_port,
        use_ssl_config=is_gateway,
        use_assets_config=is_gateway,
        use_services_configs=is_gateway,
        resolver=resolver,
        auth=auth,
        api_configs=None,
        api_location_configs=api_location_configs,
        is_local_streams_service=True,
    )
    return clean_config(config)



import logging

from celery import Task

_logger = logging.getLogger("haupt.tasks")


class HauptTask(Task):
    """Base custom celery task with basic logging."""

    abstract = True

    def on_success(self, retval, task_id, args, kwargs):
        _logger.info("Async task succeeded", extra={"task name": self.name})

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        extra = {
            "task name": self.name,
            "task id": task_id,
            "task args": args,
            "task kwargs": kwargs,
        }
        _logger.error("Async Task Failed", exc_info=einfo, extra=extra)

    def on_retry(self, exc, task_id, args, kwargs, einfo):
        _logger.info("Async task retry", extra={"task name": self.name})



import re
import uuid

import django.contrib.postgres.fields
import django.utils.timezone

from django.db import migrations, models

from polyaxon import schemas


def migrate_is_managed_by(apps, schema_editor):
    Run = apps.get_model("db", "Run")
    Run.objects.filter(is_managed=True).update(managed_by=schemas.ManagedBy.AGENT.value)
    Run.objects.filter(is_managed=False).update(managed_by=schemas.ManagedBy.USER.value)


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0012_alter_artifact_updated_at_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="run",
            name="managed_by",
            field=models.CharField(
                blank=True,
                choices=[("user", "user"), ("cli", "cli"), ("agent", "agent")],
                db_index=True,
                max_length=5,
                default=schemas.ManagedBy["AGENT"],
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="is_managed",
            field=models.BooleanField(
                default=True,
                help_text="If this entity is managed by the platform (deprecated).",
            ),
        ),
        migrations.AddField(
            model_name="run",
            name="component_state",
            field=models.UUIDField(blank=True, db_index=True, null=True),
        ),
        migrations.AddField(
            model_name="run",
            name="checked_at",
            field=models.DateTimeField(
                blank=True, db_index=True, default=django.utils.timezone.now, null=True
            ),
        ),
        migrations.AddField(
            model_name="run",
            name="controller",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="controller_runs",
                to="db.run",
            ),
        ),
        migrations.AddField(
            model_name="run",
            name="cost",
            field=models.FloatField(blank=True, default=0, null=True),
        ),
        migrations.AddField(
            model_name="run",
            name="cpu",
            field=models.FloatField(blank=True, default=0, null=True),
        ),
        migrations.AddField(
            model_name="run",
            name="custom",
            field=models.FloatField(blank=True, default=0, null=True),
        ),
        migrations.AddField(
            model_name="run",
            name="gpu",
            field=models.FloatField(blank=True, default=0, null=True),
        ),
        migrations.AddField(
            model_name="run",
            name="memory",
            field=models.FloatField(blank=True, default=0, null=True),
        ),
        migrations.AddField(
            model_name="run",
            name="schedule_at",
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
        migrations.AddField(
            model_name="run",
            name="state",
            field=models.UUIDField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="name",
            field=models.CharField(db_index=True, max_length=256),
        ),
        migrations.CreateModel(
            name="ProjectVersion",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                ("state", models.UUIDField(blank=True, db_index=True, null=True)),
                ("description", models.TextField(blank=True, null=True)),
                (
                    "tags",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=64),
                        blank=True,
                        null=True,
                        size=None,
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("readme", models.TextField(blank=True, null=True)),
                (
                    "stage",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("testing", "testing"),
                            ("staging", "staging"),
                            ("production", "production"),
                            ("disabled", "disabled"),
                        ],
                        db_index=True,
                        default=schemas.V1Stages["TESTING"],
                        max_length=16,
                        null=True,
                    ),
                ),
                (
                    "stage_conditions",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        null=True,
                    ),
                ),
                (
                    "kind",
                    models.CharField(
                        choices=[
                            ("component", "component"),
                            ("model", "model"),
                            ("artifact", "artifact"),
                        ],
                        db_index=True,
                        max_length=12,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        max_length=128,
                        validators=[
                            django.core.validators.RegexValidator(
                                re.compile("^[-a-zA-Z0-9_.]+\\Z"),
                                "Enter a valid “value” consisting of letters, numbers, underscores, hyphens, or dots.",
                                "invalid",
                            )
                        ],
                    ),
                ),
                (
                    "content",
                    models.TextField(
                        blank=True,
                        help_text="The yaml/json content/metadata.",
                        null=True,
                    ),
                ),
                (
                    "lineage",
                    models.ManyToManyField(
                        blank=True, related_name="versions", to="db.artifactlineage"
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="versions",
                        to="db.project",
                    ),
                ),
                (
                    "run",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="versions",
                        to="db.run",
                    ),
                ),
            ],
            options={
                "db_table": "db_projectversion",
                "abstract": False,
                "indexes": [
                    models.Index(fields=["name"], name="db_projectv_name_3ea950_idx")
                ],
                "unique_together": {("project", "name", "kind")},
            },
        ),
        migrations.CreateModel(
            name="RunEdge",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("values", models.JSONField(blank=True, null=True)),
                (
                    "kind",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("action", "action"),
                            ("event", "event"),
                            ("hook", "hook"),
                            ("dag", "dag"),
                            ("join", "join"),
                            ("run", "run"),
                            ("tb", "tb"),
                            ("build", "build"),
                        ],
                        db_index=True,
                        max_length=6,
                        null=True,
                    ),
                ),
                (
                    "statuses",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(
                            choices=[
                                ("created", "created"),
                                ("resuming", "resuming"),
                                ("on_schedule", "on_schedule"),
                                ("compiled", "compiled"),
                                ("queued", "queued"),
                                ("scheduled", "scheduled"),
                                ("starting", "starting"),
                                ("running", "running"),
                                ("processing", "processing"),
                                ("stopping", "stopping"),
                                ("failed", "failed"),
                                ("stopped", "stopped"),
                                ("succeeded", "succeeded"),
                                ("skipped", "skipped"),
                                ("warning", "warning"),
                                ("unschedulable", "unschedulable"),
                                ("upstream_failed", "upstream_failed"),
                                ("retrying", "retrying"),
                                ("unknown", "unknown"),
                                ("done", "done"),
                            ],
                            max_length=16,
                        ),
                        blank=True,
                        null=True,
                        size=None,
                    ),
                ),
                (
                    "downstream",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="upstream_edges",
                        to="db.run",
                    ),
                ),
                (
                    "upstream",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="downstream_edges",
                        to="db.run",
                    ),
                ),
            ],
            options={
                "db_table": "db_runedge",
            },
        ),
        migrations.AddField(
            model_name="run",
            name="upstream_runs",
            field=models.ManyToManyField(
                blank=True,
                related_name="downstream_runs",
                through="db.RunEdge",
                to="db.run",
            ),
        ),
        migrations.CreateModel(
            name="Bookmark",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                ("enabled", models.BooleanField(default=True)),
                ("object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "bookmark",
                "verbose_name_plural": "bookmarks",
                "db_table": "db_bookmark",
                "abstract": False,
            },
        ),
        migrations.RunPython(migrate_is_managed_by),
    ]



from haupt.common import auditor
from haupt.common.events.registry import artifact_version

auditor.subscribe(artifact_version.ArtifactVersionDeletedEvent)
auditor.subscribe(artifact_version.ArtifactVersionCreatedActorEvent)
auditor.subscribe(artifact_version.ArtifactVersionUpdatedActorEvent)
auditor.subscribe(artifact_version.ArtifactVersionViewedActorEvent)
auditor.subscribe(artifact_version.ArtifactVersionDeletedActorEvent)
auditor.subscribe(artifact_version.ArtifactVersionTransferredActorEvent)
auditor.subscribe(artifact_version.ArtifactVersionNewStageEvent)
